##jmd & sp
##11/2007
##chipchipnorm.R
##package to norm chip-chip data from peng's 2007 bmc bioinfo paper

#mad divided by root 2
lagNoise <- function(MA, max.lag=25, chip=1, group=NULL){
    if (!is.numeric(MA$M)){ stop("'MA' must have numeric matrix component 'M'. See vignette for details") }
    MA$M <- as.matrix(MA$M)
    if (!(max.lag %in% 1:(nrow(MA$M)-1))){ stop("'max.lag' must be an integer between 1 and ", nrow(MA$M)-1) }
    if (!(chip %in% 1:ncol(MA$M))){ stop("'chip' must be an integer between 1 and ", ncol(MA$M)) }
    if (!is.null(group) & length(group)!=nrow(MA$M)){ stop("'group' must be NULL or have length = ", nrow(MA$M)) }
    M <- as.matrix(MA$M)[,chip]
    #group
    if (!is.null(group)) {
        grp <- factor(group)
        mat <- matrix(nrow=max.lag, ncol=nlevels(grp)+1, dimnames=list(1:max.lag, 
        paste(c(levels(grp), "all"), "mad", sep=".")))
        for (i in 1:max.lag){ 
            for (j in 1:nlevels(grp)){
                mat[i,j] <- mad(diff(x=M[grp==levels(grp)[j]], lag=i))/sqrt(2)
            }#end for j
            mat[i, ncol(mat)] <- mad(diff(x=M, lag=i))/sqrt(2)
        }#end for i
        #plot
        all.colors <- rainbow(ncol(mat))
        ptColors <- rep(all.colors, each=max.lag)
        plot(x=rep(1:max.lag, times=ncol(mat)), y=mat, xlab="probe lag", ylab="deviation of lagged differences", 
        main=colnames(MA$M)[chip], col=ptColors)
        legend(x="bottomright", legend=c(levels(grp), "all"), fill=all.colors)
    }#end if
    #no group
    else { 
        mat <- matrix(nrow=max.lag, ncol=1, dimnames=list(1:max.lag, "mad"))
        #populate
        for (i in 1:max.lag){ mat[i,1] <- mad(diff(x=M, lag=i))/sqrt(2) }#end for i
        #plot
        plot(x=1:max.lag, y=mat, xlab="probe lag", ylab="deviation of lagged differences", main=colnames(MA$M)[chip])
    }#end else
    return(mat)
}#end fcn

MvAplot <- function(MA, main=NULL, xlab="A (intensity)", ylab="M (log ratio)", group=NULL, chip=1, ...){
    stopifnot(is.numeric(MA$M), is.numeric(MA$A), dim(MA$M)==dim(MA$A), chip %in% 1:ncol(MA$M))
    #already coerced M & A to matrices
    M <- MA$M[,chip]
    A <- MA$A[,chip]
    #set colors
    if (!is.null(group)){
        stopifnot(length(group)==length(M))
        group <- factor(group)
        all.colors <- rainbow(nlevels(group))
        ptColors <- all.colors[as.numeric(group)]
    }
    else { ptColors <- rep("purple", length(M)) }
    #plot
    plot(x=A, y=M, col=ptColors, xlab=xlab, ylab=ylab, main=main, ...)
    if (!is.null(group)){ legend(x="bottomleft", legend=levels(group), fill=all.colors) }
    abline(h=0, lty=2, lw=2)
}

#uses distance of (A, M) from center of mass to assess weight
#weights are 1 or 0
setWeightsMA <- function(MA, ma.out=2){
    stopifnot(is.numeric(MA$M), is.numeric(MA$A), is.numeric(ma.out))
    MA$M <- as.matrix(MA$M)
    MA$A <- as.matrix(MA$A)
    stopifnot(dim(MA$M)==dim(MA$A))
    w <- matrix(ncol=ncol(MA$A), nrow=nrow(MA$A))
    for (i in 1:ncol(MA$M)){
        M <- MA$M[,i]; m.M <- median(M, na.rm=TRUE); mad.M <- mad(M, na.rm=TRUE)
        A <- MA$A[,i]; m.A <- median(A, na.rm=TRUE); mad.A <- mad(A, na.rm=TRUE)
        #print(c("mad.M", mad.M))
        #print(c("mad.A", mad.A))
        mad.ave <- (mad.M+mad.A)/2
        M0 <- (M-m.M)/mad.ave
        A0 <- (A-m.A)/mad.ave
        d2 <- M0^2+A0^2
        if (all(d2>ma.out^2)){ stop("no points are within distance 'ma.out' for chip ", i) }
        w[,i] <- as.numeric(d2<=ma.out^2)
        #length(w)
    }#end for
    #set any NA weights to 0
    w[is.na(w)] <- 0
    return(w)
}#end fcn

#accepts vector of log-ratios and returns vector of weights in {0, 1}
#modified to reflect paper on 11.27.07 w/ shouyong's permission
setWeightsLR <- function(M, lr.out=2){
    stopifnot(is.numeric(M), is.numeric(lr.out))
    n <- length(M)
    w <- rep(1,n)
    m <- median(M, na.rm=TRUE)
    delta <- sd(M, na.rm=TRUE)
    w[(M > (m+lr.out*delta)) | (M < (m-lr.out*delta))] <- 0
    w[is.na(w)] <- 0
    return(w)
}    

#angle in radians
getAngle <- function(MA, nlag=8, w=1, w.cut=0.5, ma.out=2){
    if (!is.numeric(MA$M)){ stop("'MA' must have numeric matrix component 'M'. See vignette for details") }
    if (!is.numeric(MA$A)){ stop("'MA' must have numeric matrix component 'A'. See vignette for details") }
    MA$M <- as.matrix(MA$M)
    MA$A <- as.matrix(MA$A)
    if (!all(dim(MA$M)==dim(MA$A))){ stop("'MA$M' and 'MA$A' must have same dimension") }
    if (!(nlag %in% 1:(nrow(MA$M)-1))){ stop("'nlag' must be an integer between 1 and ", nrow(MA$M)-1) }
    if (!all(w>=0, w<=1)){ stop("all 'w' must be between 0 and 1") }
    if (!is.numeric(w.cut)|w.cut>=1){ stop("'w.cut' must be numeric below 1") }
    if (!is.numeric(ma.out)){ stop("'ma.out' must be numeric") }
    
    ##messages
    #message("Using nlag=", nlag, ". You may use lagNoise() to asses an alternate lag quantity.")
    #warn re: ordering
    #message("It is assumed (but not verified) that probes are ordered by genomic location.")
    #make matrices of differences - difference operations are carried out on each column separately.
    
    ##diffMA
    diff.m <- diff(MA$M, lag=nlag)
    diff.a <- diff(MA$A, lag=nlag)
    diffMA <- list(M=diff.m, A=diff.a)
    
    ##get weights for rotation
    #NA's propogate in diff(), so make w<w.cut to NA to exclude poor spots
    w.mat <- matrix(w, nrow=nrow(MA$A), ncol=ncol(MA$A))
    w.mat[w.mat<=w.cut] <- NA
    if (all(is.na(w.mat))){ stop("no probes pass have weight greater than 'w.cut'") }
    diff.w.initial <- diff(x=w.mat, lag=nlag)
    #setWeightsMA
    diff.w <- setWeightsMA(MA=diffMA, ma.out=ma.out)
    #combine weights
    diff.w[is.na(diff.w.initial)] <- 0
    
    ##compute angle
    angle <- numeric(ncol(diffMA$M))
    for (i in 1:ncol(diffMA$M)){
        cov.mat <- cov(x=cbind(diffMA$A[,i], diffMA$M[,i])[diff.w[,i]>w.cut,], use="complete.obs") #akin to na.rm=TRUE
        svd.c <- svd(x=cov.mat)
        #svd.c
        angle[i] <- asin(svd.c$u[1,2])
    }#end for
  return(angle)
}

#must accept full MA-object; users may want to use this fcn in workflow
rotateMA <- function(MA, angle){
    if (!is.numeric(MA$M)){ stop("'MA' must have numeric matrix component 'M'. See vignette for details") }
    if (!is.numeric(MA$A)){ stop("'MA' must have numeric matrix component 'A'. See vignette for details") }
    MA$M <- as.matrix(MA$M)
    MA$A <- as.matrix(MA$A)
    if (!all(dim(MA$M)==dim(MA$A))){ stop("'MA$M' and 'MA$A' must have same dimension") }
    if (!is.numeric(angle)){ stop("'angle' must be numeric") }
    MA0 <- MA.s <- MA.r <- MA
    for (i in 1:ncol(MA$M)){
        #o <- c(median(MA0$A), median(MA0$M))  # center of rotation
        # shift to center of rotation
        MA.s$M[,i] <- MA0$M[,i]-median(MA0$M[,i], na.rm=TRUE)
        MA.s$A[,i] <- MA0$A[,i]-median(MA0$A[,i], na.rm=TRUE)
        # rotate
        rotate.mat <- matrix(c(cos(angle[i]), sin(angle[i]), -sin(angle[i]), cos(angle[i])), nrow=2, byrow=FALSE)
        rotate.res <- rotate.mat %*% rbind(MA.s$A[,i], MA.s$M[,i])
        MA.r$A[,i] <- rotate.res[1,]
        MA.r$M[,i] <- rotate.res[2,]
        # shift back
        MA0$M[,i] <- MA.r$M[,i]+median(MA0$M[,i], na.rm=TRUE)
        MA0$A[,i] <- MA.r$A[,i]+median(MA0$A[,i], na.rm=TRUE)
    }#end for
    return(MA0)
}

#order of M & A (according to genomic loc) is essential
#incorporate weights from input MA object
lagNorm <- function(MA, nlag=8, group=NULL, w=1, w.cut=0.5, lr.out=2, ma.out=2, npoints=6000,
                    plot.names=paste("MAplots_", colnames(MA$M), sep=""),...){
  if (!is.numeric(MA$M)){ stop("'MA' must have numeric matrix component 'M'. See vignette for details") }
  if (!is.numeric(MA$A)){ stop("'MA' must have numeric matrix component 'A'. See vignette for details") }
  MA$M <- as.matrix(MA$M)
  MA$A <- as.matrix(MA$A)
  if (!all(dim(MA$M)==dim(MA$A))){ stop("'MA$M' and 'MA$A' must have same dimension") }
  if (!(nlag %in% 1:(nrow(MA$M)-1))){ stop("'nlag' must be an integer between 1 and ", nrow(MA$M)-1) }
  if (!all(w>=0, w<=1)){ stop("all 'w' must be between 0 and 1") }
  if (!is.numeric(w.cut)|w.cut>=1){ stop("'w.cut' must be numeric below 1") }
  if (!is.numeric(lr.out)){ stop("'lr.out' must be numeric") }
  if (!is.numeric(ma.out)){ stop("'ma.out' must be numeric") }
  if (!is.numeric(npoints)|npoints<=0){ stop("'npoints' must be a numeric quantity greater than 0") }
  if (!is.null(plot.names) & length(plot.names)!=ncol(MA$M)){ 
    stop("'plot.names' must be NULL or have length = ", ncol(MA$M)) 
  }
    
  #weights matrix
  w.mat <- matrix(as.numeric(w), nrow=nrow(MA$M), ncol=ncol(MA$M))
    
    #get angle
    angle <- getAngle(MA=MA, nlag=nlag, w=w.mat, w.cut=w.cut, ma.out=ma.out)
    #fit rotation to MA
    MA.r <- rotateMA(MA=MA, angle=angle)
    
    ##lowess
    MA.rl <- MA.r
    #loop through arrays
    for (i in 1:ncol(MA$M)){
        #weights
        w.temp <- setWeightsLR(M=MA.r$M[,i], lr.out=lr.out)
        if (all(w.temp==0)){ stop("no points are within distance 'lr.out' for chip ", i) }
        w.mat[w.temp==0, i] <- 0
        #fit
        loess.fit <- loess(MA.r$M[,i] ~ MA.r$A[,i], span = 0.3, degree = 1, surface = "interpolate", 
        family = "symmetric", trace.hat = "approximate", iterations = 4, weights=w.mat[,i])
        #assign
        nona.ind <- rowSums(is.na(cbind(MA.r$M[,i], MA.r$A[,i])))==0
        MA.rl$M[nona.ind, i] <- MA.r$M[nona.ind, i]-loess.fit$fitted
        #make sure if only A was NA, now M is also, since can't fit. this shouldn't happen, but user gives MA.
        MA.rl$M[!nona.ind, i] <- NA
    }#end for
    MA.norm <- MA.rl #object to return, in case need to subset MA.rl for plotting
        
    ##plot
    if (!is.null(plot.names)){
        #sample if necessary
        if (npoints<nrow(MA$M)) {
            message(npoints, " points were randomly chosen for plotting. Set 'npoints' at or above ", nrow(MA$M), 
            " to plot all points.")
            ind <- sample(x=1:nrow(MA$M), size=npoints)
            #subset
            group <- group[ind]
            MA$M <- MA$M[ind,]; MA$A <- MA$A[ind,]
            MA.r$M <- MA.r$M[ind,]; MA.r$A <- MA.r$A[ind,]
            MA.rl$M <- MA.rl$M[ind,]; MA.rl$A <- MA.rl$A[ind,]
        }#end if
        #get diffMA for plot (b)
        diffMA <- list(M=diff(MA$M, lag=nlag), A=diff(MA$A, lag=nlag))
        #loop
        for (i in 1:ncol(MA$M)){
            png(paste(plot.names[i], ".png", sep=""), height=960, width=960)
            par(mfrow=c(2,2))
            #a
            MvAplot(MA=MA, chip=i, main="pre-normalization", group=group, ...)
            #b: if group NULL, it remains NULL after removing terms 1:nlag
            MvAplot(MA=diffMA, chip=i, xlab=expression(sigma(A)), ylab=expression(sigma(M)), main="lagged differences",
            sub="brown is symmetrized fit", group=group[-(1:nlag)], ...)
            abline(a=0, b=-tan(angle[i]), lty=2, col="brown")
            #c
            MvAplot(MA=MA.r, chip=i, main="post-rotation", sub="brown is loess fit", group=group, ...)
            points(x=MA.r$A[,i], y=MA.r$M[,i]-MA.rl$M[,i], col="brown", pch=19) #loess=MA.r$M-MA.rl$M
            #d
            MvAplot(MA=MA.rl, chip=i, main="post-normalization", group=group, ...)
            abline(h=0, col="brown", lw=2)
            dev.off()        
        }#end for
        message("The following PNG images of MA plots were sent to '", getwd(), "':\n", 
        paste(plot.names, ".png \n", sep=""))
    }#end if
    return(MA.norm)
}#end fcn
